<?php

namespace Drupal\editoria11y\Form;

use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;

/**
 * Class Editoria11ySettings.
 */
class Editoria11ySettings extends ConfigFormBase {

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'editoria11y_form_settings';
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return [
      'editoria11y.settings',
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    $config = $this->config('editoria11y.settings');
    $form['setup'] = array(
      '#type' => 'fieldset',
      '#title' => t('Basic configuration'),
    );
    $form['setup']['content_root'] = [
      '#title' => $this->t("Custom content root(s)"),
      '#type' => 'textfield',
      '#placeholder' => 'main',
      '#description' => $this->t('Defaults to <code>main</code> if available. <br>Provide a custom container (or set of containers) as a comma-separated list of jQuery selectors, e.g.: <code>main, #footer</code>'),
      '#default_value' => $config->get('content_root'),
    ];
    $form['setup']['no_load'] = [
      '#title' => $this->t("Don't run if these elements are on the page"),
      '#type' => 'textfield',
      '#placeholder' => '#layout-builder',
      '#description' => $this->t('Prevents scanning the currently displayed content, e.g. while inline editing tools are open.<br>Provide a comma-separated list of jQuery selectors that should disable scanning, e.g.: <code>#my-inline-page-editor-form, .my-content-type-with-no-editable-content</code>.<br>Also consider <a href="https://www.drupal.org/project/issues/editoria11y?categories=All">submitting an issue</a> if this is due to a module conflict.'),
      '#default_value' => $config->get('no_load'),
    ];
    $form['custom'] = array(
      '#type' => 'fieldset',
      '#title' => t('Customize tests'),
    );
    $form['custom']['ignore_containers'] = [
      '#title' => $this->t("Ignore these containers"),
      '#type' => 'textfield',
      '#placeholder' => '#toolbar-administration, .contextual-region > nav',
      '#description' => $this->t('Get rid of conflicts or false positives. Some Drupal elements (<code>#toolbar-administration, .contextual-region > nav</code>) are always ignored.<br>Provide a comma-separated list of jQuery selectors for containers the tests should skip, e.g.: <code>.my-embedded-feed, #my-social-link-block</code>'),
      '#default_value' => $config->get('ignore_containers'),
    ];
    $form['custom']['embedded_content_warning'] = [
      '#title' => $this->t("Warn that these elements need manual review"),
      '#type' => 'textfield',
      '#description' => $this->t('Highlight these elements during a "Full Check."<br>Provide a comma-separated list of jQuery selectors for elements with potentially complex issues or that have been set to ignore in the previous field, e.g.: <code>.my-embedded-feed, #my-social-link-block</code>'),
      '#default_value' => $config->get('embedded_content_warning'),
    ];
    $form['text'] = array(
      '#markup' => '<p><strong>Note:</strong> test after customizing anything on this page! Changing the defaults can dramatically affect performance, and/or prevent all tests from running. If things do break, please empty these fields to restore the defaults before <a href="https://www.drupal.org/project/issues/editoria11y?categories=All">submitting an issue</a>.</p>',
    );

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $this->config('editoria11y.settings')
      ->set('ignore_containers', $form_state->getValue('ignore_containers'))
      ->set('no_load', $form_state->getValue('no_load'))
      ->set('content_root', $form_state->getValue('content_root'))
      ->set('embedded_content_warning', $form_state->getValue('embedded_content_warning'))
      ->save();
    parent::submitForm($form, $form_state);
  }

}
